/*
 * @(#)%Name% %Version% %Time%
 * 
 * Copyright (c) 1997 Rutgers University
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that (i) the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation. (ii) 
 * the name of Rutgers University may not be used in any advertising
 * or publicity relating to the software without the specific, prior
 * written permission of Rutgers University.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL RUTGERS UNIVERSITY BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * CopyrightVersion %Version%
 * @author		%Author%
 * @version		%Version%, %Time%
 * @since		ACT %Version%
 *
 */

package EDU.rutgers.CORBA.GIOP.messages;

import EDU.rutgers.act.endpoint.Endpoint;
import EDU.rutgers.act.protocol.Message;

import EDU.rutgers.act.presentation.TypeReader;
import EDU.rutgers.act.presentation.TypeWriter;
import EDU.rutgers.act.presentation.TransferSyntax;

import EDU.rutgers.act.endpoint.Endpoint;

import EDU.rutgers.CORBA.presentation.TransferSyntaxFactory;

/**
 * GIOPMessage represents the GIOP MessageHeader. It is the base class of other GIOP messages.
 * [CORBA 2.0 Specification 12-16]
 *
 * @see GIOPRequestMessage
 * @see GIOPReplyMessage
 * @see GIOPCancelRequestMessage
 * @see GIOPLocateRequestMessage
 * @see GIOPLocateReplyMessage
 * @see GIOPCloseConnectionMessage
 * @see GIOPMessageErrorMessage
 *
 * @since ACT 0.1
 */
public class GIOPMessage implements Message
{
    /**
     * GIOP magic code
     *
     * @since ACT 0.1
     */
	public final static char[] GIOPMagic =
	    {'G', 'I', 'O', 'P',};

    /**
     * Default and self generated GIOP messages are little endian
     *
     * @since ACT 0.1
     */
    public final static boolean MyByteOrder = false;


    /**
     * Used to associate the reply with request. 
     *
     * @since ACT 0.1
     */
    public int request_id_;

    /**
     * Network size of the MessaageHeader
     *
     * @since ACT 0.1
     */
    protected int giopHeaderSize_; 

    /** Indentifies the GIOP messages
     *
     * @since ACT 0.1
     */
	public char[] magic_ = GIOPMagic;

    /**
     * Versiob number of the GIOP protocol being used in the message.
     * This is not equivalent to the IIOP Version although they have the same structures.
     *
     * @since ACT 0.1
     */
    public Version GIOP_version_ = Version.Version_1_0;

    /**
     * Indicate the byte ordering used in the subsequence elements of the message 
     * (including the message_size). A value of false indicates big-endian byte ordering,
     * and true indicates little-endian byte ordering
     *
     * @since ACT 0.1
     */
	public boolean byte_order_ = MyByteOrder;

    /**
     * Indicates the type of the message.
     * Although it is enum type, it is encoded as octet
     *
     * @see MessageType
     * 
     * @since ACT 0.1
     */
	public MessageType message_type_;


    /**
     * Contains the message following the MessageHeader in octet.
     * This count includes any alignment gaps.
     * Different to other fields, it is protected.
     *
     * @see #setMessageSize
     * 
     * @since ACT 0.1
     */
	protected int message_size_ = 0;

    /**
     * The transfer syntax used in this message.
     * The default is CDR transfer syntax, the user could change this property 
     * by getting the corresponding one from the TransferSyntaxFactory. 
     * Later when write or read from the stream, the user should use the same kind
     * of transfer syntax as this one.
     *
     * @see EDU.rutgers.CORBA.presentation.TransferSyntaxFactory
     * @see EDU.rutgers.CORBA.presentation.TransferSyntaxFactory#getTransferSyntax
     *
     * @since ACT 0.1
     */
    public TransferSyntax ts_ = TransferSyntaxFactory.getTransferSyntax("CDR");
 
    /**
     * Empty constructor, it could only be used by the GIOPMessageFactory
     *
     * @since ACT 0.1
     */
    protected GIOPMessage() {
        calGIOPHeaderSize();
    }

    /**
     * Constructs the corresponding GIOP messages.
     *
     * @param MessageType type of message to be constructed
     *
     * @since ACT 0.1
     */
    protected GIOPMessage(MessageType message_type) {
		message_type_ = message_type;

        calGIOPHeaderSize();
    }

    private void calGIOPHeaderSize() {
        giopHeaderSize_ = 0;
        giopHeaderSize_ += ts_.calAlignGap(giopHeaderSize_, 4) + 4; // magic_

        // GIOP_version_    (2 bytes)
        // byte_order_      (1 byte)
        // message_type_    (1 byte)
        giopHeaderSize_ += (ts_.calAlignGap(giopHeaderSize_, 1) + 1) << 2; 

        giopHeaderSize_ += ts_.calAlignGap(giopHeaderSize_, 4) + 4;  // message_size_

    }

    /**
     * If the message is other than CloseConnection, MessageError, this field must be set.
     *
     * @param int MessageHeader message_size field
     *
     * @see #message_size_
     *
     * @since ACT 0.1
     */
    protected final void setMessageSize(int message_size) {
		message_size_ = message_size;
    }

    /**
     * Return the corresponding message SUB type. (Required by francu@caip.rutgers.edu)
     *
     * @since ACT 0.1
     */
    public int getType() {
        return message_type_.value();
    }

    private Endpoint endpoint_ = null;

    /**
     * Set the corresponding message endpoint. (Required by francu@caip.rutgers.edu)
     *
     * @since ACT 0.1
     */
    public void setEndpoint(Endpoint ep) {
        endpoint_ = ep;
    }

    /**
     * Return the corresponding message endpoint. (Required by francu@caip.rutgers.edu)
     *
     * @since ACT 0.1
     */
    public Endpoint getEndpoint() {
        return endpoint_;
    }

    /**
     * Read from the TypeReader stream. It only reads the message_size field,
     * the GIOPMessagefactory would read other fields.
     * The TypeReader should have the same transfer syntax as the ts_ member variable.
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see #ts_
     * @see ORG.omg.CORBA.Streamable#read
     * @see EDU.rutgers.act.presentation.TypeReader
     *
     * @since ACT 0.1
     */
    public void read(TypeReader in) 
        throws java.io.IOException {
        // TypeReader should be set properly
		message_size_ = in.readLong();
    }
    
    /**
     * Write to the TypeWriter stream.
     * The TypeWriter should have the same transfer syntax as the ts_ member variable.
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see #ts_
     * @see ORG.omg.CORBA.Streamable#write
     * @see EDU.rutgers.act.presentation.TypeWriter
     *
     * @since ACT 0.1
     */
    public void write(TypeWriter out)
      throws java.io.IOException {
        // reset TypeWriter
        out.setIndex(0);
        out.setByteOrder(byte_order_);

        // Header
		out.writeCharArray(magic_, 0, 4);
		GIOP_version_.write(out);
		out.writeBoolean(byte_order_);
		out.writeOctet((byte)(message_type_.value()));
		out.writeULong(message_size_);
    }
    

// Object
/*
    public boolean equals(Object o) {
		if((o != null) && (o instanceof GIOPMessage)) {
            if (o == this) {
                return true;
            }
			GIOPMessage msg = (GIOPMessage)o;
			String h1 = String.valueOf(magic_);
			String h2 = String.valueOf(msg.magic_); 
			if( h1.equals(h2) &&
				GIOP_version_.equals(msg.GIOP_version_) &&
				message_type_.equals(msg.message_type_) &&
				message_size_ == msg.message_size_) {
				return true;
			}
		}
		return false;
    }
*/

    /**
     * Returns a string representation of the GIOPMessage. 
     *
     * @return String GIOPMessage description
     *
     * @since ACT 0.1
     */
    public String toString() {
        StringBuffer buf = new StringBuffer("GIOP:GIOPMessage").
            append("\n\t[Header]").
            append("\n\tmagic: ").append(magic_).
            append("\n\tGIOP_version: ").append(GIOP_version_.toString()).
            append("\n\tmessage_type: ").append(message_type_.toString()).
            append("\n\tmessage_size: ").append(message_size_);
        return String.valueOf(buf);
	}

  public int getTransactionId()
    {
      return request_id_;
    }
}
