/*
 * @(#)%Name% %Version% %Time%
 * 
 * Copyright (c) 1997 Rutgers University
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that (i) the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation. (ii) 
 * the name of Rutgers University may not be used in any advertising
 * or publicity relating to the software without the specific, prior
 * written permission of Rutgers University.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL RUTGERS UNIVERSITY BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * CopyrightVersion %Version%
 * @author		%Author%
 * @version		%Version%, %Time%
 * @since		ACT %Version%
 *
 */

package EDU.rutgers.CORBA.GIOP.messages;

import ORG.omg.CORBA.MARSHAL;
import ORG.omg.CORBA.BAD_PARAM;

import EDU.rutgers.act.presentation.TypeReader;

import EDU.rutgers.act.protocol.Message;
import EDU.rutgers.act.protocol.MessageFactory;

/**
 * To create a GIOP message from the stream, we need to get the GIOP MessageHeader to determine
 * what kind of message should be created. After the corresponding message has been constructed,
 * that message's read method is invoked to fillin its member variables.
 *
 * @see GIOPMessage
 *
 * @see GIOPRequestMessage
 * @see GIOPReplyMessage
 * @see GIOPCancelRequestMessage
 * @see GIOPLocateRequestMessage
 * @see GIOPLocateReplyMessage
 * @see GIOPCloseConnectionMessage
 * @see GIOPMessageErrorMessage
 * 
 * @since ACT 0.1
 */
public final class GIOPMessageFactory implements MessageFactory
{
    /**
     * Create the GIOP messages from the TypeReader stream
     *
     * @param TypeReader the reader used to read from
     * 
     * @exception MARSHAL if any error occurs such as wrong magic, wrong version, etc.
     *
     * @see ORG.omg.CORBA.MARSHAL
     *
     * @since ACT 0.1
     */
    public Message createFromStream(TypeReader in) 
        throws MARSHAL {
        // reset
        in.setIndex(0);

        GIOPMessage msg = null;
        try {
		    char[] magic = new char[4];
		    in.readCharArray(magic, 0, 4);
            if (magic[0] != GIOPMessage.GIOPMagic[0] ||
                magic[1] != GIOPMessage.GIOPMagic[1] ||
                magic[2] != GIOPMessage.GIOPMagic[2] ||
                magic[3] != GIOPMessage.GIOPMagic[3]) {
                throw new MARSHAL();
            }

            Version version= new Version();
            version.read(in);
            if (version.equals(Version.Version_1_0) == false) {
                throw new MARSHAL();
            }
		
		    boolean byteOrder = in.readBoolean(); // throws MARSHAL
		    in.setByteOrder(byteOrder);

		    MessageType msgType = MessageType.type(in.readOctet()); // throws BAD_PARAM

		    switch (msgType.value()) {
			    case MessageType.REQUEST : {
				    msg = new GIOPRequestMessage();
				    break;
			    }

			    case MessageType.REPLY: {
				    msg = new GIOPReplyMessage();
				    break;
			    }
				
    			case MessageType.LOCATE_REQUEST : {
	    			msg = new GIOPLocateRequestMessage();
			    	break;
			    }

			    case MessageType.LOCATE_REPLY : {
				    msg = new GIOPLocateReplyMessage();
				    break;
			    }

                case MessageType.CANCEL_REQUEST : {
				    msg = new GIOPCancelRequestMessage();
				    break;
                }

                case MessageType.CLOSE_CONNECTION : {
				    msg = new GIOPCloseConnectionMessage();
				    break;
                }

			    case MessageType.MESSAGE_ERROR : {
				    msg = new GIOPMessageErrorMessage();
				    break;
			    }

                default : {
                    // only for the compiler
				    break;
			    }
		    }

		    msg.byte_order_ = byteOrder;
            msg.read(in);
        } catch (java.io.IOException e) {
       	    System.out.println("exception: " + e);
            // rethrows MARSHAL
            throw new MARSHAL();
        } catch (BAD_PARAM e) {
            // rethrows MARSHAL
            throw new MARSHAL();
        }
		return msg;
    }
}    
